

set -ex



test -f $PREFIX/include/arrow/api.h
test -f $PREFIX/include/arrow/flight/types.h
test -f $PREFIX/include/plasma/client.h
test -f $PREFIX/include/parquet/api/reader.h
test -f $PREFIX/lib/libarrow.so
test -f $PREFIX/lib/libarrow_flight.so
test -f $PREFIX/lib/libarrow_python.so
test -f $PREFIX/lib/libparquet.so
test -f $PREFIX/lib/libplasma.so
test ! -f $PREFIX/lib/libarrow.a
test ! -f $PREFIX/lib/libarrow_flight.a
test ! -f $PREFIX/lib/libarrow_python.a
test ! -f $PREFIX/lib/libplasma.a
test ! -f $PREFIX/lib/libparquet.a
conda inspect linkages -p $PREFIX $PKG_NAME
exit 0
